#!/bin/sh

#__________________________________________________________________________________________________
# author    : venkat@unitrends.com
# created : Fri Feb 28 12:58 EDT    
# purpose : Database Health Alert
#                     determine state change of the database and raise
#                     alert if necessary.
#Bug#        Date        Who Description
#------- ------- --- -----------------------------------------------------------------
#18040     14Mar14 VSK 

"exec" "python" "-u" "-Wignore" "$0" "$@"

import sys, os        
if not '/usr/bp/lib/monitoring' in sys.path:
	sys.path.append('/usr/bp/lib/monitoring')
import time
from datetime import datetime
import dbmon_Util as pgdb 

## globals
nc_numConn = { "nMinor":220, "nMajor":270, "nCritical":330 } 
nc_trapConf = { 0:False, 1:False, 2:False, 3:True }
nc_notifyId = { 3:170, 2:169, 1:168 } 
nc_trapType = 17

def proc_cmdline():
    """
    command line parser
    input: none
    output: options, parsed arguments, handle
    """
    from optparse import OptionParser
    curdir = os.getcwd()
    usage = "%prog [options] arg"
    parser = OptionParser(usage)
    parser.add_option("-l", "--logPath", type="string", dest="logPathFile", default='/var/log/dbmon_pgConn.log', 
            help="path+filename to dbhealth monitoring log [default=%default]")
    parser.add_option("-i", "--nMinor", type="int", dest="nMinor", default=220, 
            help="minor severity max connection limit [default=%default]")
    parser.add_option("-j", "--nMajor", type="int", dest="nMajor", default=270, 
            help="major severity max connection limit [default=%default]")
    parser.add_option("-k", "--nCritical", type="int", dest="nCritical", default=330, 
            help="critical severity max connection limit [default=%default]")
    parser.add_option("--sMinor", dest="sMinor", action="store_true", default=False, 
            help="send trap for minor severity [default=%default]")
    parser.add_option("--sMajor", dest="sMajor", action="store_true", default=False, 
            help="send trap for major severity [default=%default]")
    parser.add_option("--sCritical", dest="sCritical", action="store_true", default=True, 
            help="send trap for critical severity [default=%default]")
    parser.add_option("-d", "--durationDays", type="float", dest="nDurationDays", default=30, 
            help="duration between successive traps for given severity [default=%default days]")
    (options, args) = parser.parse_args()
    usr_args = []
    joboptions = []
    if len(args) > 0:
        for arg in args:
            if arg[0] != "-":
                joboptions.append(arg)
            else:
                usr_args.append(arg)
    return (options,args,parser)

def getSeverity(conn_now):
    """
    get severity level for trap
    input: number of connections
    output: severity level
    """

    if conn_now >= nc_numConn['nCritical']:
        sev_now = 3
    elif conn_now >= nc_numConn['nMajor']:
        sev_now = 2
    elif conn_now >= nc_numConn['nMinor']:
        sev_now = 1
    else:
        sev_now = 0
    return sev_now

def getNumConnTuple():
    """
    count all postgres processes except postmaster and *[r pro]cess (logger, writer etc.)
    ignore any "You have new mail in /var/spool/mail/root" messages
    input: none
    output: (utc-timestamp, num of db connections, severity) tuple
    """

    ## UTC timestamp now 
    tc = pgdb.TimeConvert()
    ts = tc.UTC2TimeStamp(datetime.utcnow())

    ## count number of connections to db
    cmd='ps --no-header -o pid,cmd -u postgres | egrep -v "*r pro|-D|spool" | wc -l'
    out, err, p, res = pgdb.runCommand(cmd)
    nconn=-1
    if err == '' and p.returncode == 0:
        nconn=int(out.strip())

    ## get severity
    sev = getSeverity(nconn)
    
    return (ts, nconn, sev)

def getPrevTuple(file, sev):
    """
    get closest (in time) trap info for a severity level
    input: log file, severity level
    output: (progID, utc-timestamp, num of db connections, severity) tuple or None
    """
    
    tuple = None
    
    cmd='tac %s | grep "sev=%d" | head -n 1' % (file, sev)
    out, err, p, res = pgdb.runCommand(cmd)
    if err == '' and p.returncode == 0:
        tuple = out.split()
    
    return tuple

def getKeyVals(ntuple, delimeter='='):
    """
    get key/value pairs from a tuple
    input: tuple, delimeter
    output: dict object with key/val pairs
    """
    
    result={}
    ## loop over tuple and get lv/rv pairs
    for pair in ntuple:
        value=None
        if len(pair.split(delimeter)) > 1:
            (key,value) = pair.split(delimeter)
            if value != None and key:
                result[key] = value
    
    return result

def getCmdOptions(options):
    """
    parse and store cmdline options in global vars
    input: options
    output none
    """
    items = [ "nMinor", "nMajor", "nCritical" ]
    for it in items:
        if options.nMinor and it == "nMinor":
            nc_numConn[it] = options.nMinor
            nc_trapConf[1] = options.sMinor
        elif options.nMajor and it == "nMajor":
            nc_numConn[it] = options.nMajor
            nc_trapConf[2] = options.sMajor
        elif options.nCritical and it == "nCritical":
            nc_numConn[it] = options.nCritical
            nc_trapConf[3] = options.sCritical

def getLockStatus(lockFile):
    """
    check if we have a lockfile
    input: lockfile name
    output: True/False
    """
    fileExists = os.path.isfile(lockFile)
    last_modified = None
    if fileExists:
        last_modified = os.stat(lockFile).st_mtime
    return (fileExists, last_modified)

def createLock(lockFile, severity):
    """
    create a lockfile
    input: lockfile name, severity level
    output: status (True -- ALL OK, False -- ERROR)
    """
    status = False
    if os.path.exists(lockFile):
        #os.utime(lockFile, None)
        return status
    else:
        try:
            open(lockFile,'w').close()
            status = True
        except IOError:
            pass
    return status

def send(message, objectID, severity, configFile, setOpen, notification_id = -1):
    """
    send snmp trap using monitoring framework

	string message  - the message to include with the trap
	string objectID - the object ID to include with the trap
	string configFile - the path the to the cached version of the database
	bool setOpen - whether or not to close the trap
    """
    send_status = False
    try:
        trap_params = pgdb.getTrapParams(configFile, severity)
        if trap_params and len(trap_params) >= pgdb.SNMPTables().numItems:
            send_status = pgdb.sendTrap(message, objectID, trap_params, setOpen)
    except:
        pass

    return send_status

def dbConn(options, args, parser):
    """
    main
    """
    progID = str(os.path.splitext(os.path.basename(__file__))[0])
    
    ## get/load cmdline options
    getCmdOptions(options)

    ## load defaults
    configFile=os.path.join('/usr/bp/schema','.dbmon_tableDump')
    logPathFile=os.path.join('/var/log', 'dbmon_'+progID+'.log')
    lockFileStub=os.path.join('/var/lock/subsys', '.'+progID)

    ## load cmdline options
    if options.logPathFile:
        logPathFile = options.logPathFile

    nDurationDays = 30
    if options.nDurationDays:
        nDurationDays = options.nDurationDays

    minDuration_sec = float(nDurationDays * 24 * 3600)
 
    ## define trap message format
    msgformat = '%s ts=%.2f sev=%d nc=%d tp=%d'

    ## get the number of connections to db and log it
    tims_now, conn_now, sev_now = getNumConnTuple()
    if conn_now == -1:
        sys.exit(0)
    
    ## should we send a trap for this severity (default = no)
    trap_now = 0
    
    ## unique identifier for traps
    objectID = progID

    ## no previous tuple (or we don't know if there was one)
    ## check lockfile. if it exists, don't send it. 
    ## for the moment lock overrides everything else
    lockFile = '%s.%d' % (lockFileStub, int(sev_now))

    locked, fModtim = getLockStatus(lockFile)

    ## if locked and elapsed    time >    min duration for this severity, remove lock 
    removeLock = False
    if locked:
        removeLock = (tims_now - fModtim) > minDuration_sec and (sev_now > 0)
        if removeLock:
            try:
                os.remove(lockFile)
            except:
                pass

    ## get previous tuple for this severity (if we can) AND
    ## if there was a previous trap sent get elapsed time
    prev_tuple = getPrevTuple(logPathFile, sev_now)
    if prev_tuple:
        kv_tuple = getKeyVals(prev_tuple)
        if kv_tuple:
            tims_prev = float(kv_tuple['ts'])
            elaps_time = (tims_now - tims_prev)
            if elaps_time > minDuration_sec and nc_trapConf[sev_now] and sev_now > 0:
                trap_now = 1
    else:
        if not locked and nc_trapConf[sev_now] and sev_now > 0:
            trap_now = 1

    ## compose trap message and log it
    message_now = msgformat % (progID, tims_now, sev_now, conn_now, trap_now)
    #trap_now=1 ## testing
    
    ## send trap
    if trap_now == 1:
        if createLock(lockFile, sev_now):
            trapRaise = True
            send(message_now, objectID, sev_now, configFile, trapRaise)
        ## write and close log
        log = pgdb.setupLogger(logPathFile)
        if log:
            try:
                log.write(message_now+'\n')
                log.close()
            except IOError:
                pass

#__________________________________________________________________________________________________
## py conditional script hook
if __name__ == "__main__":
    options, args,parser = proc_cmdline()
    sys.exit(dbConn(options,args,parser))



