__name__ = 'Cron'
__doc__  = 'A package for the manipulation of crontab files'
class cronTask:
	"""
	Class that store the crontab Task

	"""

	def getMinute(self):
		"""
		Returns the Minute that the script is to run
		"""
		return self.freq[0]

	def getHour(self):
		"""
		Returns the Hour that the script is to run
		"""
		return self.freq[1]

	def getDay(self):
		"""
		Returns the Day that the script is to run
		"""
		return self.freq[2]

	def getMonth(self):
		"""
		Returns the Month that the script is to run
		"""
		return self.freq[3]

	def getDOW(self):
		"""
		Returns the Day Of the Week that the script is to run
		see crontab for the more information
		"""
		return self.freq[4]

	def __init__(self, id, freq, command, user = "root"):
		"""
		Create a cron object
		string id - name of the cron task
		string[5] freq - frequency of the task
		string command - task to be run
		string user - user to run the script
		"""
		self.id = id
		self.freq = freq
		self.command = command
		self.user = user
		if command == "":
			raise TypeError('Expected a non-empty string')
		if isinstance(self.freq,str):
			self.freq = freq.split()
		if isinstance(self.command,str):
			self.command = command.split()
		if isinstance(self.freq,list) and len(self.freq) != 5:
			raise TypeError("Expected a list of lenght 5")

	def __str__(self):
		"""
		Prints the string in crontab format
		"""
		return "#ID %s\n%s %s %s" % \
				(self.id, ' '.join(self.freq),self.user, ' '.join(self.command))
		
class cronManager:
	"""
	Class that manages the cronFile

	values from cronFilePath are not loaded by default; call load to read the file

	Changes are NOT commited until the write() function is called
	"""

	def __init__(self, cronFilePath, **kwargs):
		"""
		Create the cronManager Object ; if values are still not set use some
		defaults

		Keyword Arguments:
		stirng cronFilePath - path to the file where the crontab is stored
		dict **kwargs - a dictionary that map envioment varibles to to there values
		"""
		self.cronFilePath = cronFilePath
		self.tasks = {}
		self.varibles = dict([(key.upper(),value) for (key,value) in kwargs.iteritems()])
		# get the values from the files
		# use the passed values if they were passed

		#if the varibles are not set yet use the defaults
		try:
			self.varibles['PATH']
		except KeyError:
			self.varibles['PATH'] = "/sbin:/usr/sbin:/bin:/usr/bin:/usr/local/bin:/usr/bp/bin:/usr/bp/lib/shlib:/root/bin:/usr/local/pgsql/bin"

		try:
			self.varibles['SHELL']
		except KeyError:
			self.varibles['SHELL'] = "/bin/bash"
		
		try:
			self.varibles['MAILTO']
		except KeyError:
			self.varibles['MAILTO'] = "root"

		try:
			self.varibles['HOME']
		except KeyError:
			self.varibles['HOME'] = "/"

		try:
			self.varibles['PYTHONPATH']
		except KeyError:
			self.varibles['PYTHONPATH'] = "/usr/bp/lib/monitoring:/usr/bp/bin"

	def reload(self):
		"""
		Reloads the configuration file
		"""
		self.tasks.clear()
		self.load()

	def load (self):
		"""
		Loads the settings from the crontab overwriting the stored varibles

		Throws: DuplicateJobError if there are more than 1 script with the same
			id in the same file returns: none
		Throws: IOError if the cronFile does not exist
		"""
		cronFile = open(self.cronFilePath,'r')
		id = ""
		for line in cronFile:
			if id:
				freq = line.strip().split()[:5]
				command = line.strip().split()[6:]
				user = line.strip().split()[5]
				self.insert(id,freq,command,user)
				id = ""
				continue
			try:
				if line.strip()[0:3] == '#ID' :
					id = line.strip()[4:].strip()
					continue
			except IndexError:
				#This can happen when user modifies the config file manually
				continue
			value = line.strip().split('=')
			if len(value) == 2:
				self.varibles[value[0].strip().upper()] = value[1].strip()
		
	def find (self, id):
		"""
		returns a cron object with the specified id or None if not found

		keyword arguments:
		string id - id value to search for

		returns: cronTask or None if the task does not exist
		"""
		return self.tasks.get(id)
	
	def insert (self, id, freq, command, user='root'):
		"""
		Insert a new task into the crontab

		Keyword Arguments
		string id - Unique key that identifies the cron tab
		string[5] freq - a string that specifies the cron frequency
			see man crontab for more details
		string command[] - string that specifies the command to be run
		string user - user to execute the task

		Throws: DuplicateJobError if the id is a duplicate
		"""
		if self.tasks.has_key(id):
			raise DuplicateJobError('The id %s already exists'%id)
		else:
			self.tasks[id] = cronTask(id,freq,command,user)

	def remove (self, id):
		"""
		Remove a task from the crontab

		Keyword Arguments
		string id - Unique key that identifies the cron tab

		Throws: NonExistantJobError if id is not in the crontab
		Returns: the cronTask that was removed
		"""
		if not self.tasks.has_key(id):
			raise NonExistantJobError("%s does not exist"% id)
		else:
			return self.tasks.pop(id)

	def update(self, id, freq="", command="",user = ""):
		"""
		Overwrites the freq and command for the task with 'id'
		Any unprovided arguments will be preserved

		Keyword Arguments
		string id - Unique key that identifies the cron tab
		string[5] freq - a string that specifies the cron frequency
			see man crontab for more details
		string command - specifies the command to be run
		string user - specifies the user to run the command

		Throws: NonExistantJobError if id is not in the crontab
		"""
		tmp = self.remove(id)
		if isinstance(freq, str):
			freq = freq.split()
		if isinstance(command, str):
			command = command.split()
		if freq == []:
			freq = tmp.freq
		if command == []:
			command = tmp.command
		if user == "":
			user = tmp.user
		self.insert(id, freq, command, user)

	def write (self):
		"""
		Overwrites the varible of the class to the crontab
		"""
		cronFile = open(self.cronFilePath,'w')
		cronFile.write(str(self))
		cronFile.close()
	
	def __iter__(self):
		"""
		iterator the the cronManager class

		Returns: an iterator to all of the tasks
		"""
		return self.tasks.iteritems()

	def __str__(self):
		"""
		Prints all of the tasks

		Returns: string representation of the cronfile
		"""
		string = ""
		for varible, value in self.varibles.iteritems():
			string += varible + ' = ' + value + '\n'
		for id, task in self.tasks.iteritems():
			string += str(task) + '\n'
		return string

class cronError(Exception):
	"""BaseError for all cronManager errors"""
class NonExistantJobError(cronError):
	"""Error that indicates that the desired id is not in the crontab"""
class DuplicateJobError(cronError):
	"""Error that indicates that the desired id is already in the crontab"""
	
