#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# author	: Venkat Kaushik <venkat at unitrends.com>
# created : MAR 2014
# purpose : Database Health Monitoring Utilities 
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

import os, sys
import time
import datetime


class SNMPTables:
	"""
	Enumeration of tables used in SNMP trap
	"""
	numItems = 5
	notifications, categories, trap_types, nvp, snmp_config	= range(numItems)
	name = [ 'notifications', 'categories', 'trap_types', 'nvp', 'snmp_config' ]
	uniqID = [ 'notifID_', 'catID_', 'trapType_', 'nvp_', 'trapServ_' ]
	
class Platform(object):
	"""
	# Get Unitrends Rapid Recovery platform specific information
	"""

	def getHostName(self):
		"""
		Gets the network hostname

		Keyword Arguments:
		None

		Returns:
		String hostname 
		"""
		import socket
		return socket.gethostname()

	def getMACAddr(self):
		"""
		Get MAC address of primary (default/gateway) network interface

		Keyword Arguments:
		None

		Returns:
		string MAC Address or None if there was an error
		"""

		cmd = '/sbin/route | awk \'{if($1 ~/defa/ && $4 ~/G/) print $NF;}\''
		ifID = None
		self.macID='00:00:00:00:00:00'
		## get interface ID
		out, err, p, res = runCommand(cmd)
		if err == '' and p.returncode == 0 and out:
			ifID = out.strip()

		## check if interface ID is available
		cmd = 'test -d /sys/class/net && find /sys/class/net -maxdepth 1 -name %s ' % ifID
		cmd += '-type d -exec basename {} \; | head -n 1'
		out, err, p, res = runCommand(cmd)
		if err == '' and p.returncode == 0:
			cmd = '/sbin/ifconfig %s | grep -i HWaddr | awk \'{print $NF}\'' % ifID
			out, err, p, res = runCommand(cmd)
			if err == '' and p.returncode == 0 and out:
				self.macID = out.strip()
		return self.macID

	def getAssetTag(self, uassetFile='/etc/unitrends-asset'):
		"""
		Get DPU Asset Tag (NNN-NN-NNNNN)

		Keyword Arguements:
		Path uassetFile - path to the unitrends asset tag file

		Returns:
		string - returns the asset tag and '000-000-0000' if one is not found
		"""
		uafile = uassetFile
		assetTag = None
		if os.path.isfile(uafile):
			try:
				f = open(uafile,'r')
				if f:
					assetTag = f.readline().strip('\n')
			except:
				pass
		## if not available, then try dpu asset cmdline
		if assetTag == None:
			cmd = 'dpu asset | awk \'{print $NF}\''
			out, err, p, res = runCommand(cmd)
			if err == '' and p.returncode == 0:
				assetTag = out.strip()
			else:
				assetTag = '000-000-00000'

		return assetTag

	def getSWInstallTime(self,package='unitrends-rr.x'):
		"""
		Get install time of an rpm package

		Keyword Arguments:
		string package - package to check install time

		Return
		string swinsdate - date the package was installed or None if there was an error
		"""
		swinstdate = 'none'
		cmd = 'dpu version | grep Installed | sed -e "s/Ins.*on\ //g"'
		out, err, p, res = runCommand(cmd)
		if err == '' and p.returncode == 0 and out:
			swinstdate = out.strip()
		else:
			cmd = 'rpm -q --qf \'%{INSTALLTIME:date}\' %s | egrep -v "not installed"' % package
			if err == '' and p.returncode == 0 and out:
				swinstdate = out.strip()
		return swinstdate

	def getSWRelease(self,package='unitrends-rr.x'):
		"""
		Get latest installed version

		Keyword Arguments:
		string package - package to check install version

		Return
		string swversion - version number for the package or None if there was an error
		"""
		swversion = 'none'
		cmd = 'dpu version | grep Version | awk \'{print $2}\''
		out, err, p, res = runCommand(cmd)
		if err == '' and p.returncode == 0 and out:
			swversion = out.strip()
		else:
			cmd = 'rpm -q --qf \'%{VERSION}-%{RELEASE}\' %s | egrep -v "not installed"' % package
			if err == '' and p.returncode == 0 and out:
				swversion = out.strip()
		return swversion

class TimeConvert(object):
	"""
	Time Conversion Utils for Unitrends Monitoring
	"""
	
	import datetime
	def UTC2TimeStamp(self, dt, epoch=datetime.datetime(1970,1,1)):
		"""
		Convert UTC to timestamp
		"""
		td = dt - epoch
		timestamp = (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6)/1e6
		return timestamp

	def TimeStamp2UTC(self, ts):
		"""
		Convert timestamp to UTC
		"""
		return datetime.datetime.utcfromtimestamp(ts)

	def TimeNow(self):
		"""
		Get time (dt) now
		"""
		return datetime.datetime.now()

	def UTC2Local(self,utc):
		"""
		Convert UTC to localtime
		"""
		epoch = time.mktime(utc.timetuple())
		offset = datetime.datetime.fromtimestamp(epoch) - datetime.datetime.utcfromtimestamp(epoch)
		return utc + offset

class DBUtil(object):
	"""
	Unitrends Database Pulse Check Utils
	"""

	def isDBUp(self):
		"""
		Check if database server process is active
		
		Keyword Arguments:
		none

		Returns:
		returns 1 if the db is up and 0 if the db is down
		"""
		cmd = 'ps ax | grep [p]ostgres > /dev/null'
		out, err, p, res = runCommand(cmd)
		return err == '' and p.returncode == 0

	def pingDB(self):
		"""
		Check if database accepts query from cmdline

		Keyword Arguments:
		none

		Returns:
		returns 1 if the db accepts queries from db and 0 else
		"""
		if self.isDBUp():
			cmd = 'psql postgres postgres -Xqltc "\q" > /dev/null'
			out, err, p, res = runCommand(cmd)
			return err == '' and p.returncode == 0

def readSamples(path):
	"""
	Parse csv file and generate a dictionary object

	Keyword parameters:
	string path - read the samples from the file at path

	return:
	stirng[][] rows - information from the csv file or None if error
	"""

	import csv, StringIO
	## check if path exists
	if not os.path.isfile(path):
		return None

	## try opening path/file
	csvfile = open(path, 'rb')
	if not csvfile:
		return None

	## read data
	data = ''
	for line in csvfile.readlines():
		data += line

	## modify / instantiate dialect
	class PGDB_Dialect(csv.Dialect):
		quotechar='\x07'
		delimiter = '|'
		lineterminator = '\n'
		doublequote = False
		skipinitialspace = True
		quoting = csv.QUOTE_NONE
		escapechar = '\\'

	dbDialect=PGDB_Dialect()

	## reader initialize
	reader = csv.DictReader(StringIO.StringIO(data), dialect=dbDialect)
	
	## read and store rows
	rows = [ row for row in reader ]
	
	return rows

def setupLogger(tFullPath, max_size_bytes=308224, fopen_mode='a+'):
	"""
	returns a logger file

	Keyword Arguments:
	string tFullPath - full path to to the file to written to/read
	int max_size_bytes - files larger than this will be overwritten
	fopen_mode - mode to open the file see help(Open)

	Returns - a file object or None if there was an error
	"""
	log = None
	mode = fopen_mode
	if os.path.isfile(tFullPath):
		st = os.stat(tFullPath)
		if st.st_size >= max_size_bytes:
			mode = 'w'
	try:
		log = open(tFullPath, mode)
	except IOError:
		pass
	return log

def runCommand(cmd, rUsage=False):
	"""
	Run system command and get output, error, process handle and resource usage handle

	Keyword Arguments
	string cmd - command to be run
	bool rUsage - determines if the resource usage should be tracked

	Returns:
	string out - output from the command
	string err - stderr from the command
	Popen - pipe object from the command
	struct_rusage res - resource useage object from the command or None if error or rUsage=False
	"""
	from subprocess import Popen, PIPE, STDOUT
	from resource import getrusage,RUSAGE_SELF,RUSAGE_CHILDREN
	res = None
	out = None
	err = None
	try:
		p = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE)
		p.wait()
	except OSError:
		print err
		sys.exit(p.returncode)
	(out, err) = p.communicate()
	if rUsage:
		res = getrusage(RUSAGE_CHILDREN)
	return (out, err, p, res)

def pingHost(HostIP, Verbose=False):
	"""
	Send ICMP ECHO_REQUEST 

	Keyword Arguements:
	string HostIP - ip to ping
	Verbose - No effect; to be implemented

	Return:
	int code - 1 if sucessful, 0 if fewer replys than expected, and -1 for other error
	"""
	## note, only root can ping with i < 1 and so we restrict 
	if not (HostIP and os.geteuid()==0):
		return -1
	cmd = 'ping -s 4 -c 2 -i 0.05 -q -w 1 %s > /dev/null 2>&1; echo $?' %	HostIP
	out, err, p, res = runCommand(cmd)
	out = out.strip()
	if out == '0':
		return 1
	if out == '1':
		return 0
	else:
		return -1

def getConfigSectionMap(config, section):
	"""
	Read configuration (ConfigParser) object and convert to python dictionary

	Keyword Arguments:
	ConfigParser config - config parser to use to construct the dictionary
	sting section - section to get from the config parser

	Return:
	Returns a dictionary corresponding to the contents of the cache file
	"""
	import ConfigParser
	cfgmap = {}
	if not config:
		return cfgmap
	options = config.options(section)
	for opt in options:
		try:
			cfgmap[opt] = config.get(section, opt)
			if cfgmap[opt] == -1:
				pass
		except ConfigParser.InterpolationSyntaxError:
			cfgmap[opt] = config.get(section, opt, 1)
		except:
			cfgmap[opt] = None

	return cfgmap

def getConfigParser(iniPath):
	"""
	Get and process config (.ini) file and convert to configuration object
	
	Keyword Arguments:
	string iniPath - path to load the cached database

	Returns:
	ConfigParser or none if there was an error
	"""
	import ConfigParser
	config = None
	if not os.path.isfile(iniPath):
		return config
	try:
		config = ConfigParser.SafeConfigParser()
		config.read(iniPath)
	except:
		pass
	return config

def buildTrapParams(notification_id, category_id, trap_type_id, severity,  primary_text, short_description, action, category_name, trap_type, state, psa, alert,trap,   email, log,  configFile = '/usr/bp/schema/.dbmon_tableDump' ):
	"""
	Builds a traps from the cached database

	Keyword Parameters
	int notification_id
	int category_id
	int trap_type_id
	int severity
	string primary_text -
	string short_description - string that 
	string action - string that suggests a resolution
	string category_name - string that coresponds to the category name
	string trap_type - string that corresponds to the trap type
	char state  - 't' if true 'f' if false; not currently implemented;
	char psa	- 't' if true 'f' if false; not currently implemented;
	char alert  - 't' if true 'f' if false; not currently implemented;
	char trap   - 't' if true 'f' if false; not currently implemented;
	char email  - 't' if true 'f' if false; not currently implemented;
	char log	- 't' if true 'f' if false; not currently implemented;
	string configFile - path to the configuration file
	"""
	
	import re
	
	notifid = None
	confdata = {}

	tc = SNMPTables() 
	
	cfg = getConfigParser(configFile)
	## configurations needed for sending trap
	config_notifid = {
						'notification_id'   : str(notification_id),
						'category_id'		: str(category_id),
						'severity'			: str(severity),
						'primary_text'		: primary_text,
						'short_description' : short_description,
						'state'				: state,
						'action'			: action,
						'psa'				: psa ,
						'alert'				: alert,
						'trap'				: trap ,
						'email'				: email,
						'log'				: log
					 }
	config_catid =   {
						'category'		  : category_name,
						'category_id'	   : category_id
					 }
	config_traptype ={
						'trap_type_id'		: trap_type_id,
						'trap_type'			: trap_type
					 }
	config_hostid = None
	config_tserver = None

	## get notification id configuration, if so get category configuration
	confdata[tc.notifications] = config_notifid
	confdata[tc.categories] = config_catid
	confdata[tc.trap_types] = config_traptype

	## get host type configuration
	hostid = tc.uniqID[tc.nvp]+'Identity'
	config_hostid = getConfigSectionMap(cfg, hostid)
	if not config_hostid:
		return None
	confdata[tc.nvp] = config_hostid

	## get trap server configuration
	uniq_tsc = tc.uniqID[tc.snmp_config]
	tsc_pattern = re.compile(uniq_tsc+'[0-9]+')
	ntrap_servers = []
	for sec in cfg.sections():
		if tsc_pattern.match(sec):
			ntrap_servers.append(sec)
	for tserver in ntrap_servers:
		try:
			tconf = getConfigSectionMap(cfg, tserver)
			if not tconf:
				continue
			confdata[tserver] = tconf
		except:
			pass
	
	return confdata

def getTrapParams(configFile, severity=1, notification_id = -1):
	"""
	Gets the information for sending traps from the cached database

	Keyword Parameters
	string configFile - path to the cached database file made by dbmon_UpdateConfig
	int severity - severity for the database not pushing alerts
	int notification_id - notification id to search the database for
	"""
	
	import re
	
	notifid = None
	confdata = {}

	tc = SNMPTables()

	if severity < 1:
		return None
	cfg = getConfigParser(configFile)
	if not cfg:
		return None 

	## configurations needed for sending trap
	config_notifid = None
	config_catid = None
	config_traptype = None
	config_hostid = None
	config_tserver = None
	try:
		if (notification_id == -1):
			notifid = tc.uniqID[tc.notifications]+str(nc_notifyId[severity])
		else:
			notifid = tc.uniqID[tc.notifications]+str(notification_id)
	except:
		pass
	## get notification id configuration, if so get category configuration
	config_notifid = getConfigSectionMap(cfg, notifid)
	if not config_notifid:
		return None
	else:
		if config_notifid['category_id']:
			catid = tc.uniqID[tc.categories]+config_notifid['category_id']
			config_catid = getConfigSectionMap(cfg, catid)
			if not config_catid:
				return None
	confdata[tc.notifications] = config_notifid
	confdata[tc.categories] = config_catid

	## get trap type configuration
	if (notification_id == -1):
		ttype = tc.uniqID[tc.trap_types]+str(nc_trapType)
	else:
		ttype = tc.uniqID[tc.trap_types]+str(config_notifid['category_id'])
	config_traptype = getConfigSectionMap(cfg, ttype)
	if not config_traptype:
		return None
	confdata[tc.trap_types] = config_traptype

	## get host type configuration
	hostid = tc.uniqID[tc.nvp]+'Identity'
	config_hostid = getConfigSectionMap(cfg, hostid)
	if not config_hostid:
		return None
	confdata[tc.nvp] = config_hostid

	## get trap server configuration
	uniq_tsc = tc.uniqID[tc.snmp_config]
	tsc_pattern = re.compile(uniq_tsc+'[0-9]+')
	ntrap_servers = []
	for sec in cfg.sections():
		if tsc_pattern.match(sec):
			ntrap_servers.append(sec)
	for tserver in ntrap_servers:
		try:
			tconf = getConfigSectionMap(cfg, tserver)
			if not tconf:
				continue
			confdata[tserver] = tconf
		except:
			pass
	
	return confdata

def sendTrap(message, objectID, trap_params, setOpen=True):
	"""
	sends an SNMP trap 

	Keyword Arguments:
	string message - message to send as part of the trap
	string objectID - objectID to send as part of the trap
	string [][] trap_params - cached copy of the datbases as a dictionary
	bool setOpen - true sets the status to 1; false sets status to 0

	Returns:
	returns -1 if there was an error and > 160 if at least one sucess
	"""
	## get system info
	tc = SNMPTables()

	## OID, Type, Identifier, Message, Severity, Status, Asset/MAC, Hostname/Type, Version/InstallDate
	trp = {}
	whspace = ' '
 
	## SNMP Trap server(s)
	ntrap_servers = len(trap_params) - tc.numItems + 1
	if ntrap_servers <= 0:
		return -1
	#print "finish intial traps"
	tservrip = [None]*ntrap_servers
	tservcom = [None]*ntrap_servers
	tenabled = [False]*ntrap_servers
	for i in range(ntrap_servers):
		sec = tc.uniqID[tc.snmp_config]+str(i)
		try:
			tenabled[i] = trap_params[sec]['enabled'] == 't'
			tservrip[i] = trap_params[sec]['destination']
			tservcom[i] = trap_params[sec]['community']
			if tenabled[i]:
				if pingHost(tservrip[i]) < 1:
					tenabled[i] = False
		except:
			pass
		if not tenabled[i]:
			continue

	if not (True in tenabled):
		return -1 
	#print "finish trap server setup" 
	## OID
	catok = False
	category = None
	severity = None
	trapstate = None
	try:
		category = trap_params[tc.notifications]['category_id']
		catok = trap_params[tc.categories]['category_id'] == category
		severity = trap_params[tc.notifications]['severity']
		trapstate = trap_params[tc.notifications]['state']
	except:
		pass
	if not (category and severity and trapstate):
		return -1
	

	## Type
	traptype = None; traptypeId = None
	try:
		traptype = trap_params[tc.trap_types]['trap_type']
		traptypeId = trap_params[tc.trap_types]['trap_type_id']
	except:
		pass
	if not ( traptype and traptypeId):
		return -1
	#print "finished trap type"
	trp['type'] = traptype
	
    #print "finished trap setup"
	trp['oid'] = '0.'+traptypeId
	
	## Message
	if not message:
		return -1
	#print "finished message"
	trp['message'] = message.rstrip('\n')

	## Severity
	trp['severity'] = severity

	## State (open/close -- revisit) 1 = open, 0 = close 
	trp['state'] = str(int(setOpen))	

	## Asset Tag / MAC Address / Hostname
	assetmac = None
	macaddr = None
	try:
		p = Platform()
		assettag = p.getAssetTag()
		macaddr = p.getMACAddr()
		nnhname = p.getHostName()
		swrelease = p.getSWRelease()
		swinstdate = p.getSWInstallTime()
	except:
		pass
	if not (assettag and macaddr and nnhname and swrelease and swinstdate):
		return -1
	#print "finished source info"

	trp['assetmac'] = assettag+whspace+macaddr

	## Host Type
	hosttype = None
	try:
		hosttype = trap_params[tc.nvp]['item_value']
	except:
		pass
	if not hosttype:
		return -1
	#print "print hosttype"
	trp['hosttype'] = nnhname+whspace+hosttype

	## SW Version/ Install Date
	trp['swversion'] = swrelease+whspace+swinstdate

	
	## trap command string
	trapcmd	= 'snmptrap -v 2c -c'
	trapcmd += ' %s'													## community
	trapcmd += ' %s \'\' '										## trap-server ip
	
	## iso.org.dod.internet.private.enterprises.21865 
	##	.1.	3.	6.			 1.			4.					1.21865 followed by
	## unitrends enterprise number, product, trap group, trap number
	## http://www.unitrends.com/documents/administrators-guide/user_manual/advanced_configuration/snmp_trap_conditions.htm

	trap_prefix = { 0:'enterprises.21865.1.', 1:'enterprises.21865.1.1.', 2:'2.100.' }
	prefix = trap_prefix[1]
	payload	= trap_prefix[0] + trap_prefix[2] + trp['oid'] + whspace	 ## OID
	payload += prefix + '1 s \'' + trp['type'] + '\'' + whspace				 ## Type
	payload += prefix + '2 s \'' + objectID + '\'' + whspace						## ObjectID
	payload += prefix + '3 s \'' + message + '\'' + whspace						 ## Message
	payload += prefix + '4 i ' + trp['severity'] + whspace							## Severity
	payload += prefix + '5 i ' + trp['state'] + whspace								 ## State
	payload += prefix + '6 s \'' + trp['assetmac'] + '\'' + whspace		 ## Asset/MAC
	payload += prefix + '7 s \'' + trp['hosttype'] + '\'' + whspace		 ## HostType
	payload += prefix + '8 s \'' + trp['swversion'] + '\''						 ## Version/Install date

	retcode = 150
	for i in range(ntrap_servers):
		cmd = trapcmd % (tservcom[i],tservrip[i])
		cmd += payload
		out, err, p, res = runCommand(cmd)
		if err == '' and p.returncode == 0:
			retcode += 10
		else:
			#print err
			retcode += 1

	return retcode

